import std/lists
import std/sequtils
import std/strformat
import std/strutils
import std/math

import convenience
import types

func moveCheck*(line: string): MoveType =
  ## Checks the type of movement. F moves without X/Y movement
  ## are considered as a Z hop or layer change, and E (extrude),
  ## and F (no extrude) moves are set to extrude/travel.
  let travelCheck = line.split(" ")
  if travelCheck[^1].startsWith("F") and travelCheck.len >= 3:
    # travelCheck.len > 3 avoids Z only travels
    return MoveType.travel
  elif travelCheck[^1].startsWith("F"):
    # If X Y travel is not done, this is assumed to be a layer change
    return MoveType.layerChange
  elif travelCheck[^1].startsWith("E"):
    return MoveType.extrude

func setPosition*(x, y, z: float64, line: string): tuple[x, y, z, speed:float64] =
  ## Sets the position. Takes X, Y, and/or Z in a gcode line and will
  ## assume that they haven't changed if they aren't specified.
  var
    newX: float64
    newY: float64
    newZ: float64
    newSpeed: float64

  for i in line.split:
    let trimmed = i[1 .. ^1] # The potential float minus the letter
    if i.contains "X": newX = trimmed.parseFloat.round(5)
    if i.contains "Y": newY = trimmed.parseFloat.round(5)
    if i.contains "Z": newZ = trimmed.parseFloat.round(5)
    if i.contains("F") or i.contains("E"): newSpeed = trimmed.parseFloat.round(5)

  if newX == 0.0: newX = x
  if newY == 0.0: newY = y
  if newZ == 0.0: newZ = z

  return (newX, newY, newZ, newSpeed)

func accelFormat*(line: string): string =
  ## Transforms comment line like 'ACCEL: 8000/8000/8   for Travel'
  ## to something understood by klipper. Output will be
  ## 'SET_VELOCITY_LIMIT ACCEL=8000 ACCEL_TO_DECEL=8000 SQUARE_CORNER_VELOCITY=8'
  ## for the example shown.
  let stripped = line.split(":")[1].strip.split(" ")[0].split("/")
  if stripped.len > 2:
    let accel = stripped[0]
    let decel = stripped[1]
    let scv = stripped[2]
    return fmt"SET_VELOCITY_LIMIT ACCEL={accel} ACCEL_TO_DECEL={decel} SQUARE_CORNER_VELOCITY={scv}"
  else:
    fmt"Parsing of accels failed: {line}".debugEcho
    return ""

func parseKlipperAccel*(line: string, oldAccels: KlipperAccels): KlipperAccels =
  ## Parses accels in a standardized formats to inject. The format looks as such
  ## ; TYPE:External perimeter         ACCEL=8000 ACCEL_TO_DECEL=8000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Overhang perimeter         ACCEL=1000 ACCEL_TO_DECEL=1000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Internal perimeter         ACCEL=1000 ACCEL_TO_DECEL=1000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Top solid infill           ACCEL=1000 ACCEL_TO_DECEL=1000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Solid infill               ACCEL=2000 ACCEL_TO_DECEL=2000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Internal infill            ACCEL=4000 ACCEL_TO_DECEL=4000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Bridge infill              ACCEL=8000 ACCEL_TO_DECEL=8000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Internal bridge infill     ACCEL=8000 ACCEL_TO_DECEL=8000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Thin wall                  ACCEL=1000 ACCEL_TO_DECEL=1000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Gap fill                   ACCEL=1000 ACCEL_TO_DECEL=1000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Skirt                      ACCEL=8000 ACCEL_TO_DECEL=8000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Support material           ACCEL=1000 ACCEL_TO_DECEL=1000 SQUARE_CORNER_VELOCITY=8
  ## ; TYPE:Support material interface ACCEL=1000 ACCEL_TO_DECEL=1000 SQUARE_CORNER_VELOCITY=8
  var newAccels = oldAccels
  if "for First Layer" in line:
    newAccels.firstLayer = line.accelFormat
  elif "for Travel" in line:
    newAccels.travel = line.accelFormat
  elif "TYPE:External perimeter" in line:
    newAccels.externalPerimeter = line.accelFormat
  elif "TYPE:Overhang perimeter" in line:
    newAccels.overhangPerimeter = line.accelFormat
  elif "TYPE:Internal perimeter" in line:
    newAccels.internalPerimeter = line.accelFormat
  elif "TYPE:Top solid infill" in line:
    newAccels.topSolidInfill = line.accelFormat
  elif "TYPE:Solid infill" in line:
    newAccels.solidInfill = line.accelFormat
  elif "TYPE:Internal bridge infill" in line:
    newAccels.internalBridgeInfill = line.accelFormat
  elif "TYPE:Internal infill" in line:
    newAccels.internalInfill = line.accelFormat
  elif "TYPE:Bridge infill" in line:
    newAccels.bridgeInfill = line.accelFormat
  elif "TYPE:Thin wall" in line:
    newAccels.thinWall = line.accelFormat
  elif "TYPE:Gap fill" in line:
    newAccels.gapFill = line.accelFormat
  elif "TYPE:Skirt" in line:
    newAccels.skirt = line.accelFormat
  elif "TYPE:Support material interface" in line:
    newAccels.supportMaterialInterface = line.accelFormat
  elif "TYPE:Support material" in line:
    newAccels.supportMaterial = line.accelFormat
  elif "TYPE:Custom" in line:
    newAccels.custom = line.accelFormat
  return newAccels

func parseTemp*(line: string): float =
  ## Returns all data after S as a float in a temperature
  ## set grode line
  return line.split(" ")[1][1 .. ^1].parseFloat

proc openAndParse*(fileName: string): DoublyLinkedList[GcodeState] =
  ## Opens and fully parses a gcode file
  try:
    # Opens and parses gcode object. Returns a Gcodes object
    let gcodeFile = fileName.read_file.split("\n")

    # Initialize state
    # Initialize accels
    var accels: KlipperAccels = (
          externalPerimeter: "",
          internalPerimeter: "",
          overhangPerimeter: "",
          topSolidInfill: "",
          solidInfill: "",
          internalInfill: "",
          bridgeInfill: "",
          internalBridgeInfill: "",
          thinWall: "",
          gapFill: "",
          skirt: "",
          supportMaterial: "",
          supportMaterialInterface: "",
          custom: "",
          other: "",
          firstLayer: "",
          travel: ""
    )

    var state: GcodeState = (
          preInject: "",
          moveMode: BaseGcode.absolute,
          moveType: MoveType.travel,
          feature: Feature.other,
          layer: 0,
          posX: 0.0,
          posY: 0.0,
          posZ: 0.0,
          speed: 0.0,
          comment: "",
          parsedComment: "",
          postInject: "",
          klipperAccels: accels,
          extruderTemp: 0.0,
          bedTemp: 0.0
    )


    var allGcode = initDoublyLinkedList[GcodeState]()

    for line in gcodeFile:
      state.comment = "" # Reset comment
      state.parsedComment = "" # Reset parsed comment
      state.preInject = "" # reset injections
      state.postInject = "" # reset injections
      if line.isFeature:
        state.feature = line.stripFeature
        state.parsedComment = line

      elif line.islayerChange:
        state.layer.inc
        state.parsedComment = line

      elif line.split[0] in [$BaseGcode.move, $BaseGcode.moveAlt]:
        state.moveType = line.moveCheck
        (state.posX, state.posY, state.posZ, state.speed) = setPosition(state.posX,
            state.posY, state.posZ, line)

      elif line.split[0] == $BaseGcode.absolute:
        state.moveMode = BaseGcode.absolute
        state.parsedComment = line  # TODO remove

      elif line.split[0] == $BaseGcode.relative:
        state.moveMode = BaseGcode.relative
        state.parsedComment = line  # TODO remove

      elif line.startsWith("; ACCEL"):
        state.parsedComment = line
        state.klipperAccels = parseKlipperAccel(line, state.klipperAccels)

      elif line.split[0] == $BaseGcode.extruderTemp:
        state.extruderTemp = line.parseTemp
        state.parsedComment = line  # TODO remove

      elif line.split[0] == $BaseGcode.bedTemp:
        state.bedTemp = line.parseTemp
        state.parsedComment = line  # TODO remove

      elif line.isComment or line.isEmptyOrWhitespace:
        # If comments aren't parsed for info, they are just stored in the line state.
        state.comment = line

      elif line.split[0] in BaseGcode.mapIt($it):
        # Base gcode not implimented
        fmt"ERROR: Unimplimented gcode. Treating as comment: {line}".debugEcho
        state.comment = line

      elif line.split[0] in BlindGcode.mapIt($it):
        # Treats blind gcodes like comments
        state.comment = line

      elif line.split[0] in KlipperBlindGcode.mapIt($it):
        # Treats blind gcodes like comments
        state.comment = line

      else:
        fmt"Unsupported gcode. Treating as comment: {line}".debugEcho
        state.comment = line

      allGcode.add(state)
    return allGcode

  except:
    echo getCurrentExceptionMsg()
    fmt"Could not read {fileName} file.".echo
    quit(QuitFailure)

